"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const graphql_1 = require("graphql");
const OPERATION_MARKER = '@operation';
function validateFixtureAgainstSchema(fixture, schema) {
    const queryType = schema.getQueryType();
    const mutationType = schema.getMutationType();
    const { content, path } = fixture;
    return {
        fixturePath: path,
        validationErrors: Object.keys(content).reduce((allErrors, key) => {
            let rootType;
            const keyPath = key;
            if (key === OPERATION_MARKER) {
                return allErrors;
            }
            if (objectTypeHasFieldWithName(queryType, key)) {
                rootType = queryType;
            }
            else if (objectTypeHasFieldWithName(mutationType, key)) {
                rootType = mutationType;
            }
            else {
                return allErrors.concat([error(keyPath, 'Field does not exist on query or mutation types')]);
            }
            return allErrors.concat(validateValueAgainstType(content[key], rootType.getFields()[key].type, keyPath));
        }, []),
    };
}
exports.validateFixtureAgainstSchema = validateFixtureAgainstSchema;
function objectTypeHasFieldWithName(type, name) {
    return type.getFields()[name] != null;
}
function normalizeOperationName(operationName) {
    return operationName.replace(/(Query|Mutation|Subscription)$/i, '');
}
function validateFixtureAgainstAST(fixture, ast) {
    const fixtureDirectoryName = path_1.basename(path_1.dirname(fixture.path));
    const operationMarkerName = fixture.content[OPERATION_MARKER];
    const operationName = operationMarkerName || fixtureDirectoryName;
    const operation = ast.operations[normalizeOperationName(operationName)] || ast.operations[operationName];
    if (operation == null) {
        let lookedFor = '';
        if (operationMarkerName) {
            const normalizedOperationMarkerName = normalizeOperationName(operationMarkerName);
            lookedFor = `'${operationMarkerName}'${operationMarkerName === normalizedOperationMarkerName ? '' : ` and '${normalizedOperationMarkerName}'`} based on the '${OPERATION_MARKER}' key`;
        }
        else {
            const normalizedDirectoryName = normalizeOperationName(fixtureDirectoryName);
            lookedFor = `'${fixtureDirectoryName}'${fixtureDirectoryName === normalizedDirectoryName ? '' : ` and '${normalizedDirectoryName}'`} based on the fixture’s directory name`;
        }
        throw new Error([
            `Could not find a matching operation (looked for ${lookedFor}).`,
            `Make sure to put your fixture in a folder named the same as the operation, or add an '${OPERATION_MARKER}' key indicating the operation.`,
            `Available operations: ${Object.keys(ast.operations).join(', ')}`
        ].join(' '));
    }
    const { fields, filePath, operationType } = operation;
    const value = Object.assign({}, fixture.content);
    delete value[OPERATION_MARKER];
    return {
        fixturePath: fixture.path,
        operationName,
        operationType,
        operationPath: filePath === 'GraphQL request' ? undefined : filePath,
        validationErrors: fields.reduce((allErrors, field) => {
            return allErrors.concat(validateValueAgainstFieldDescription(value[field.responseName], field, '', ast));
        }, []),
    };
}
exports.validateFixtureAgainstAST = validateFixtureAgainstAST;
function validateValueAgainstFieldDescription(value, fieldDescription, parentKeyPath, ast) {
    const { type, responseName } = fieldDescription;
    const keyPath = updateKeyPath(parentKeyPath, responseName);
    const typeErrors = validateValueAgainstType(value, type, keyPath, { shallow: true });
    if (typeErrors.length > 0) {
        return typeErrors;
    }
    return Array.isArray(value)
        ? validateListAgainstFieldDescription(value, fieldDescription, fieldDescription.type, keyPath, ast)
        : validateValueAgainstObjectFieldDescription(value, fieldDescription, keyPath, ast);
}
function validateValueAgainstObjectFieldDescription(value, fieldDescription, keyPath, ast) {
    if (value == null || typeof value !== 'object' || Array.isArray(value)) {
        return [];
    }
    const { fields = [], fragmentSpreads = [], type } = fieldDescription;
    let fragmentFields = [];
    if (fragmentSpreads) {
        fragmentSpreads
            .map((spread) => ast.fragments[spread])
            .forEach((fragment) => {
            fragment.fields.forEach((field) => {
                if (fields.some(({ fieldName }) => fieldName === field.fieldName)) {
                    return;
                }
                const isGuaranteedTypeMatch = fragment.possibleTypes.includes(makeTypeNullable(type));
                fragmentFields.push(isGuaranteedTypeMatch
                    ? field
                    : Object.assign({}, field, { type: makeTypeNullable(field.type) }));
            });
        });
    }
    return validateValueAgainstFields(value, fields.concat(fragmentFields), keyPath, ast);
}
function makeTypeNullable(type) {
    return type instanceof graphql_1.GraphQLNonNull
        ? type.ofType
        : type;
}
function validateListAgainstFieldDescription(value, fieldDescription, type, keyPath, ast) {
    const itemType = type instanceof graphql_1.GraphQLNonNull
        ? type.ofType.ofType
        : type.ofType;
    return value.reduce((allErrors, item, index) => {
        const itemKeyPath = updateKeyPath(keyPath, index);
        const itemTypeErrors = validateValueAgainstType(item, itemType, itemKeyPath, { shallow: true });
        if (itemTypeErrors.length > 0) {
            return allErrors.concat(itemTypeErrors);
        }
        return Array.isArray(item)
            ? allErrors.concat(validateListAgainstFieldDescription(item, fieldDescription, itemType, itemKeyPath, ast))
            : allErrors.concat(validateValueAgainstObjectFieldDescription(item, fieldDescription, itemKeyPath, ast));
    }, []);
}
function validateValueAgainstFields(value, fields, keyPath, ast) {
    const finalValue = value || {};
    const excessFields = Object
        .keys(finalValue)
        .filter((key) => fields.every(({ responseName }) => key !== responseName))
        .map((key) => error(keyPath, `has key '${key}' not present in the query`));
    return fields.reduce((allErrors, field) => {
        return allErrors.concat(validateValueAgainstFieldDescription(finalValue[field.responseName], field, keyPath, ast));
    }, excessFields);
}
function validateValueAgainstType(value, type, keyPath, options = { shallow: false }) {
    const { shallow } = options;
    if (type instanceof graphql_1.GraphQLNonNull) {
        return value == null
            ? [error(keyPath, `should be non-null but was ${String(value)}`)]
            : validateValueAgainstType(value, type.ofType, keyPath, options);
    }
    if (value === null) {
        return [];
    }
    const valueType = typeof value;
    if (type instanceof graphql_1.GraphQLList) {
        if (!Array.isArray(value)) {
            return [error(keyPath, `should be an array (or null), but was ${articleForType(valueType)} ${valueType}`)];
        }
        return shallow
            ? []
            : value.reduce((allErrors, item, index) => (allErrors.concat(validateValueAgainstType(item, type.ofType, updateKeyPath(keyPath, index), options))), []);
    }
    if (value === undefined) {
        const typeName = nameForType(type);
        return [error(keyPath, `should be ${articleForType(typeName)} ${typeName} (or null), but was undefined`)];
    }
    if (type instanceof graphql_1.GraphQLObjectType) {
        if (valueType === 'object') {
            if (shallow) {
                return [];
            }
            else {
                const fields = type.getFields();
                return Object.keys(value).reduce((fieldErrors, key) => {
                    const fieldKeyPath = updateKeyPath(keyPath, key);
                    if (fields[key] != null) {
                        return fieldErrors.concat(validateValueAgainstType(value[key], fields[key].type, fieldKeyPath, options));
                    }
                    else {
                        return fieldErrors.concat([error(fieldKeyPath, `does not exist on type ${type.name} (available fields: ${Object.keys(fields).join(', ')})`)]);
                    }
                }, []);
            }
        }
        else {
            return [error(keyPath, `should be an object but was a ${valueType}`)];
        }
    }
    if (type === graphql_1.GraphQLString) {
        return valueType === 'string'
            ? []
            : [error(keyPath, `should be a string but was a ${valueType}`)];
    }
    else if (type === graphql_1.GraphQLInt) {
        if (typeof value === 'number') {
            return Number.isInteger(value)
                ? []
                : [error(keyPath, 'should be an integer but was a float')];
        }
        return [error(keyPath, `should be an integer but was a ${valueType}`)];
    }
    else if (type === graphql_1.GraphQLFloat) {
        return !Number.isNaN(value)
            ? []
            : [error(keyPath, `should be a float but was a ${valueType}`)];
    }
    else if (type === graphql_1.GraphQLBoolean) {
        return typeof value === 'boolean'
            ? []
            : [error(keyPath, `should be a boolean but was a ${valueType}`)];
    }
    if (type instanceof graphql_1.GraphQLScalarType) {
        return type.parseValue(value) != null
            ? []
            : [error(keyPath, `value does not match scalar ${nameForType(type)}`)];
    }
    if (type instanceof graphql_1.GraphQLEnumType) {
        return type.parseValue(value) != null
            ? []
            : [error(keyPath, `value does not match enum ${nameForType(type)} (available values: ${type.getValues().map((enumValue) => enumValue.value).join(', ')})`)];
    }
    return [];
}
const CUSTOM_NAMES = {
    [graphql_1.GraphQLBoolean.name]: 'boolean',
    [graphql_1.GraphQLFloat.name]: 'float',
    [graphql_1.GraphQLInt.name]: 'integer',
    [graphql_1.GraphQLString.name]: 'string',
};
function nameForType(type) {
    return CUSTOM_NAMES.hasOwnProperty(type.name) ? CUSTOM_NAMES[type.name] : type.name;
}
const TYPES_WITH_ARTICLE_AN = [
    'object',
    'integer',
    'array',
];
function articleForType(type) {
    return TYPES_WITH_ARTICLE_AN.includes(type)
        ? 'an'
        : 'a';
}
function updateKeyPath(keyPath, newKey) {
    if (typeof newKey === 'number') {
        return `${keyPath}[${newKey}]`;
    }
    return keyPath ? `${keyPath}.${newKey}` : newKey;
}
function error(keyPath, message) {
    return { keyPath, message };
}
